/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>


#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <time.h>
#include <fcntl.h>

#include <errno.h>
#include "tun.h"

#include "error.h"
#include "types.h"
#include "error.h"
#include "icmp.h"

#define nelem(x) (int)(sizeof(x)/sizeof(x[0]))

typedef struct Icmptype Icmptype;
struct Icmptype {
	int type;
	int code;
	char *tag;
	char *desc;
};

Icmptype icmptypes[] = {
	{3, 0, "unr-net", "network unreachable"},
	{3, 1, "unr-host", "host unreachable"},
	{3, 2, "unr-proto", "protocol unreachable"},
	{3, 3, "unr-port", "port unreachable"},
	{3, 4, "needfrag", "fragmentation needed and DF set"},
	{3, 5, "sr-fail", "source route failed"},

	{4, 0, "src-quench", "source quench"},

	{11, 0, "ttl-exceed", "time to live exceeded in transit"},
	{11, 1, "frag-exceed", "fragment reassembly time exceeded"},

	{12, 0, "param-prob", "parameter problem"},

	{5, 0, "redir-net", "redirect network"},
	{5, 1, "redir-host", "redirect host"},
	{5, 2, "redir-tosnet", "redirect network for tos"},
	{5, 3, "redir-toshost", "redirect host for tos"},

	{8, 0, "echo-req", "echo message"},
	{0, 0, "echo-rep", "echo reply"},

	{13, 0, "ts-req", "timestamp message"},
	{14, 0, "ts-rep", "timestamp reply"},

	{15, 0, "info-req", "information request"},
	{16, 0, "info-rep", "information reply"},

	{125, 353, "garb1", "pure garbage"},
};

int type;
int code;
int tunfd;

unsigned char buf[2048];

int
main(int argc, char *argv[])
{
	int opt;
	int i, n, nwr;
	u32int origin;
	char *p;

	setbuf(stdout, NULL);

	while((opt = getopt(argc, argv, "ln:t:c:")) != -1)
	switch(opt){
	case 'l':
		printf("the following tags are usable with %s -n <tag>\n", argv[0]);
		printf("\t%-13s %s\n", "tag", "description");
		for(i = 0; i < nelem(icmptypes); i++)
			printf("\t%-13s %s\n", icmptypes[i].tag, icmptypes[i].desc);
		exit(1);
	case 'n':
		for(i = 0; i < nelem(icmptypes); i++)
			if(!strcmp(icmptypes[i].tag, optarg)){
				type = icmptypes[i].type;
				code = icmptypes[i].code;
				break;
			}
		break;
	case 'c':
		code = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	case 't':
		type = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	default:
	usage:
		printf("usage: %s [-l] [-n name] [-t type] [-c code] host-ip4 program-ip4 net-ip4/net-mask\n", argv[0]);
		exit(1);
	}

	if(optind != argc - 3)
		goto usage;

	tunfd = opentun(argv[optind], argv[optind+1]);
	snprintf(buf, sizeof buf, "route add -net %s gw %s", argv[optind+2], argv[optind+1]);
	printf("system %s\n", buf);
	system(buf);
	origin = inet_addr(argv[optind+1]);

	while((n = read(tunfd, buf, sizeof buf)) > 0){
		printf("----\n");
		for(i = 0; i < n; i++)
			printf(" %02x", buf[i]);
		printf("\n");
		switch(type*100 + code){
		case 300: /* net unreachable */
		case 301: /* host unreachable */
		case 302: /* protocol unreachable */
		case 303: /* port unreachable */
		case 304: /* frag needed & df set */
		case 305: /* source route failed */
			nwr = icmpgen(type, code, 0,0,0,0, &origin, buf, n, sizeof buf);
			break;
		case 400: /* source quench */
			nwr = icmpgen(type, code, 0,0,0,0, &origin, buf, n, sizeof buf);
			break;
		case 500: /* redirect network */
		case 501: /* redirect host */
		case 502: /* redirect tos to network */
		case 503: /* redirect tos to host */
			nwr = icmpgen(type, code, 11,12,13,14, &origin, buf, n, sizeof buf);
			nwr = -1;
			break;
		case 1100: /* time to live exceeded in transit */
		case 1101: /* fragment reassembly time exceeded */
			nwr = icmpgen(type, code, 0,0,0,0, &origin, buf, n, sizeof buf);
			break;
		case 1200: /* parameter problem message */
			/* 1:tos field, 20:first option, etc. */
			nwr = icmpgen(type, code, 1,0,0,0, &origin, buf, n, sizeof buf);
			break;

		case 800: /* echo request */
		case   0: /* echo reply */
			nwr = -1;
			break;
		case 1300: /* timestamp msg */
		case 1400: /* timestamp reply */
			nwr = -1;
			break;
		case 1500: /* information request */
		case 1600: /* information reply */
			nwr = -1;
			break;
		default:
			nwr = -2;
			break;
		}
		switch(nwr){
		default:
			write(tunfd, buf, nwr);
			break;
		case -1:
			printf("type %d code %d unsupported\n", type, code);
			break;
		case -2:
			printf("type %d code %d entirely unknown\n", type, code);
			break;
		}
	}

	return 0;
}
