/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>


#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <time.h>
#include <fcntl.h>

#include <errno.h>

#include <poll.h>

#include "error.h"
#include "types.h"

long tobesent = 10000000; /* 10 MB */
int uflag, dflag, cflag, qflag;
int maxseg;

u64int time_start, time_end;
u64int xmit_start, xmit_end;
u64int recv_start, recv_end;

char rbuf[4096];
char wbuf[4096];

int
read_to(int fd, void *vp, int len, int to)
{
	struct pollfd pfd;
	pfd.fd = fd;
	pfd.events = POLLIN;
	switch(poll(&pfd, 1, to)){
	case -1:
		error("bad select");
	case 0:
		return -1;
	default:
		return read(fd, vp, len);
	}
}

u64int
time_us(void)
{
	u64int ms;
	struct timeval tv;
	gettimeofday(&tv, NULL);
	ms = (u64int)tv.tv_sec * 1000000;
	ms += (u64int)tv.tv_usec;
	return ms;
}

u16int
decu16(char *buf)
{
	return buf[0] + (buf[1]<<8);
}

u32int
decu32(char *buf)
{
	return buf[0] + (buf[1]<<8) + (buf[2]<<16) + (buf[3]<<24);
}

u64int
decu64(char *buf)
{
	int i;
	u64int v = 0;
	for(i = sizeof v - 1; i >= 0; i--){
		v <<= 8;
		v |= (unsigned char)buf[i];
	}
	return v;
}

void
encu16(char *buf, u16int v)
{
	buf[0] = v & 0xff;
	buf[1] = (v >> 8) & 0xff;
}

void
encu32(char *buf, u32int v)
{
	unsigned int i;
	for(i = 0; i < sizeof v; i++){
		buf[i] = v & 0xff;
		v >>= 8;
	}
}

void
encu64(char *buf, u64int v)
{
	unsigned int i;
	for(i = 0; i < sizeof v; i++){
		buf[i] = v & 0xff;
		v >>= 8;
	}
}

int
main(int argc, char *argv[])
{
	int n;
	int fd;
	int opt;
	struct sockaddr_in sa;
	int port;
	char *ipaddr, *p;
	long nrecv;
	long left;

	//setbuf(stdout, NULL);

	while((opt = getopt(argc, argv, "m:b:udcq")) != -1)
	switch(opt){
	case 'q':
		qflag++;
		break;
	case 'm':
		maxseg = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	case 'b':
		tobesent = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	case 'u':
		uflag++;
		break;
	case 'd':
		dflag++;
		break;
	case 'c':
		cflag++;
		break;
	default:
	usage:
		printf("usage: %s [-ud] [-q] [-b nbytes] [-m maxseg] [-c] srv-ip4-addr port\n", argv[0]);
		exit(1);
	}

	if(optind != argc - 2)
		goto usage;

	ipaddr = argv[optind++];
	port = strtol(argv[optind++], NULL, 10);

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);

	if((fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
		error("socket");
	if(maxseg != 0)
		setsockopt(fd, IPPROTO_TCP, TCP_MAXSEG, &maxseg, sizeof maxseg);
	if(cflag){
		if(connect(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
			error("connect");
	} else {
		int opt = 1;
		socklen_t len;
		int nfd;
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt);
		if(bind(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
			error("bind");
		if(listen(fd, 1) == -1)
			error("listen");
		len = sizeof sa;
		if((nfd = accept(fd, (struct sockaddr*)&sa, &len)) == -1){
			printf("accept %s\n", strerror(errno));
			return 1;
		}
		close(fd);
		fd = nfd;
	}
	time_start = time_us();

	nrecv = 0;
	if(uflag && !dflag){
		left = tobesent;
		encu16(wbuf, sizeof wbuf);
		while(tobesent == -1 || left > sizeof wbuf - 2){
			encu64(wbuf + 2, time_us());
			if(write(fd, wbuf, sizeof wbuf) == -1)
				error("write");
			left -= sizeof wbuf;
		}
		encu16(wbuf, left);
		if(left >= 10)
			encu64(wbuf + 2, time_us());
		if(write(fd, wbuf, left) == -1)
			error("write");
		shutdown(fd, SHUT_WR);
		while((n = read_to(fd, wbuf, sizeof wbuf, 300*1000)) > 0)
			;
	} else if(!uflag && dflag){
		unsigned int need;
		unsigned int nrd = 0;
		while((n = read_to(fd, rbuf + nrd, sizeof rbuf - nrd, 300*1000)) > 0){
			nrd += n;
			nrecv += n;
			if(nrd >= 2 && nrd >= (need = decu16(rbuf))){
				if(need >= 10){
					u64int ts = decu64(rbuf + 2);
					if(qflag == 0)
						printf("times %lu %lld\n", nrecv, time_us() - ts);
				}
				memmove(rbuf, rbuf + need, nrd - need);
				nrd -= need;
			}
		}
		if(n == -1)
			error("read timeout");
	} else if(uflag && dflag){
		fd_set rset, wset;
		int rdone = 0, wdone = 0;
		unsigned int need;
		unsigned int nrd = 0;
		left = tobesent;
		while(!rdone || !wdone){
			FD_ZERO(&rset);
			FD_ZERO(&wset);
			if(rdone == 0)
				FD_SET(fd, &rset);
			if(wdone == 0)
				FD_SET(fd, &wset);
			select(fd+1, &rset, &wset, NULL, NULL);
			if(FD_ISSET(fd, &rset)){
				n = read(fd, rbuf + nrd, sizeof rbuf - nrd);
				if(n == -1)
					error("read");
				if(n == 0)
					rdone = 1;
				else {
					recv_end = time_us();
					if(recv_start == 0)
						recv_start = recv_end;
					nrd += n;
					nrecv += n;
					if(nrd >= 2 && nrd >= (need = decu16(rbuf))){
						if(need >= 10){
							u64int ts = decu64(rbuf + 2);
							if(qflag == 0)
								printf("times %lu %lld\n", nrecv, recv_end - ts);
						}
						memmove(rbuf, rbuf + need, nrd - need);
						nrd -= need;
					}
				}
			}
			if(FD_ISSET(fd, &wset)){
				encu16(wbuf, sizeof wbuf);
				if(left > sizeof wbuf - 2){
					encu64(wbuf + 2, time_us());
					if(write(fd, wbuf, sizeof wbuf) == -1)
						goto wfinish;
					xmit_end = time_us();
					if(xmit_start == 0)
						xmit_start = xmit_end;
					left -= sizeof wbuf;
				} else {
					encu16(wbuf, left);
					if(left >= 10)
						encu64(wbuf + 2, time_us());
					write(fd, wbuf, left);
					xmit_end = time_us();
					if(xmit_start == 0)
						xmit_start = xmit_end;
wfinish:
					shutdown(fd, SHUT_WR);
					wdone = 1;
				}
			}
		}
	}
	close(fd);
	time_end = time_us();

	if(qflag < 2){ 
		fd = creat("data", 0666);
		if(uflag && dflag){
			u64int rtmp, wtmp;
			rtmp = (nrecv * 8 * 1000ull) / (recv_end - recv_start);
			wtmp = ((tobesent-left) * 8 * 1000ull) / (xmit_end - xmit_start);
			snprintf(wbuf, sizeof wbuf, "%s %d 31 %llu.%03llu %s-(bytes*8)/microsec %llu.%03llu %s-(bytes*8)/microsec\n",
				inet_ntoa(sa.sin_addr), port,
				wtmp / 1000ull, wtmp % 1000ull, "upload",
				rtmp / 1000ull, rtmp % 1000ull, "download");
		} else {
			u64int tmp;
			if(dflag)
				tmp = (nrecv * 8 * 1000ull) / (time_end - time_start);
			else
				tmp = ((tobesent-left) * 8 * 1000ull) / (time_end - time_start);
			snprintf(wbuf, sizeof wbuf, "%s %d 31 %llu.%03llu %s-(bytes*8)/microsec\n",
				inet_ntoa(sa.sin_addr), port, tmp / 1000ull, tmp % 1000ull, 
				uflag ? "upload" : "download");
		}
		write(fd, wbuf, strlen(wbuf));
		close(fd);
	}

	return 0;
}

