/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>


#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include "error.h"

#define nelem(x) (int)(sizeof(x)/sizeof(x[0]))

typedef struct Proto Proto;
struct Proto {
	char *tag;
	int type;
	int protocol;
};

#ifndef IPPROTO_DCCP
#define IPPROTO_DCCP 33
#endif

#ifndef TCP_RUBBISH
#define TCP_RUBBISH 15
#endif

Proto protos[] = {
	{ "tcp", SOCK_STREAM, IPPROTO_TCP },
	{ "udp", SOCK_DGRAM, IPPROTO_UDP },
	{ "sctp", SOCK_STREAM, IPPROTO_SCTP },
	{ "sctp-sp", SOCK_SEQPACKET, IPPROTO_SCTP },
	{ "dccp", 6, IPPROTO_DCCP },
};

int rflag;
int xflag;
int cflag;
int mode;

char roptbuf[32];
int roptlen = 2;

char buf[512];
fd_set rset, wset;

int termin, termout;
int netin, netout;

void
sigalrm(int sig)
{
	sig=sig; /* stupid gcc */
}

int
main(int argc, char *argv[])
{
	int i;
	int fd;
	int opt;
	struct sockaddr_in sa;
	int port;
	char *ipaddr;
	char *p;
	Proto *proto;

	setbuf(stdout, NULL);

	proto = protos; /* tcp */
	while((opt = getopt(argc, argv, "p:cm:r:x:")) != -1)
	switch(opt){
	case 'x':
		xflag++;
		rflag = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	case 'r':
		rflag = strtol(optarg, &p, 10);
		if(*p == ':'){
			int byte;
			do {
				byte = strtol(p+1, &p, 10);
				roptbuf[roptlen++] = byte;
			} while(p != NULL && *p == '.');
		} else if(*p != '\0') {
			goto usage;
		} else for(i = 2; i < 7; i++)
			roptbuf[roptlen++] = rand();
		break;
	case 'm':
		mode = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	case 'c':
		cflag++;
		break;
	case 'p':
		for(i = 0; i < nelem(protos); i++){
			if(!strcmp(protos[i].tag, optarg)){
				proto = protos + i;
				break;
			}
		}
		break;
	default:
	usage:
		printf("usage: %s [-c] [-p protocol] srv-ip4-addr port\n", argv[0]);
		printf("\t-c connects instead of listening\n");
		printf("\t-p is one of: %s", protos[0].tag);
		for(i = 1; i < nelem(protos); i++){
			printf(", %s", protos[i].tag);
		}
		printf("\n");
		exit(1);
	}

	if(optind != argc - 2)
		goto usage;

	ipaddr = argv[optind++];
	port = strtol(argv[optind++], NULL, 10);

	if(xflag){
		struct sockaddr_in xa;
		socklen_t xlen = sizeof xa;;

		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = inet_addr(ipaddr);
		sa.sin_port = htons(port);
		if((fd = socket(PF_INET, proto->type, proto->protocol)) == -1)
			error("socket");
		if(cflag){
			if(connect(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
				error("connect");
			getsockname(fd, (struct sockaddr*)&xa, &xlen);
		} else {
			int opt = 1;
			socklen_t len;
			int nfd;
			setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt);
			if(bind(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
				error("bind");
			if(proto->protocol != IPPROTO_UDP){
				if(listen(fd, 1) == -1)
					error("listen");
				len = sizeof sa;
				if((nfd = accept(fd, (struct sockaddr*)&sa, &len)) == -1){
					printf("accept %s\n", strerror(errno));
					return 1;
				}
				close(fd);
				fd = nfd;
			}
			getpeername(fd, (struct sockaddr*)&xa, &xlen);
		}
		close(fd);
		memcpy(roptbuf + 2, &xa.sin_addr, sizeof xa.sin_addr);
		memcpy(roptbuf + 2 + sizeof xa.sin_addr, &xa.sin_port, sizeof xa.sin_port);
		roptlen = 2 + sizeof xa.sin_addr + sizeof xa.sin_port;
	}

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);

	if((fd = socket(PF_INET, proto->type, proto->protocol)) == -1)
		error("socket");

	if(rflag && proto->type == SOCK_STREAM && proto->protocol == IPPROTO_TCP){
		roptbuf[0] = rflag;
		roptbuf[1] = roptlen;
		setsockopt(fd, SOL_TCP, TCP_RUBBISH, &roptbuf, roptlen); 
	}

	if(cflag){
		sleep(1); /* ugly hack ... */
		if(connect(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
			error("connect");
	} else {
		int opt = 1;
		socklen_t len;
		int nfd;
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt);
		if(bind(fd, (struct sockaddr*)&sa, sizeof sa) == -1)
			error("bind");
		if(proto->protocol != IPPROTO_UDP){
			if(listen(fd, 1) == -1)
				error("listen");
			len = sizeof sa;
			if((nfd = accept(fd, (struct sockaddr*)&sa, &len)) == -1){
				printf("accept %s\n", strerror(errno));
				return 1;
			}
			close(fd);
			fd = nfd;
		}
	}

	if(mode == 0)
	while(netin != -1 || termin != -1){
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		if(netin == 0)
			FD_SET(fd, &rset);
		if(netout == 0)
			FD_SET(fd, &wset);
		if(termin == 0)
			FD_SET(0, &rset);
		if(termout == 0)
			FD_SET(1, &wset);
		select(fd+1, &rset, &wset, NULL, NULL);
		if(FD_ISSET(fd, &rset))
			netin = 1;
		if(FD_ISSET(fd, &wset))
			netout = 1;
		if(FD_ISSET(0, &rset))
			termin = 1;
		if(FD_ISSET(1, &wset))
			termout = 1;
		if(termin == 1 && netout == 1){
			int n;
			if((n = read(0, buf, sizeof buf)) > 0){
				write(fd, buf, n);
				termin = 0;
				netout = 0;
			} else {
				termin = -1;
				netout = -1;
				close(0);
				shutdown(fd, SHUT_WR);
			}
		}
		if(netin == 1 && termout == 1){
			int n;
			if((n = read(fd, buf, sizeof buf)) > 0){
				write(1, buf, n);
				netin = 0;
				termout = 0;
			} else {
				netin = -1;
				termout = -1;
				shutdown(fd, SHUT_RD);
				close(1);
			}
		}
	}

	if(mode == 1){
		struct sigaction act;
		memset(&act, 0, sizeof act);
		act.sa_handler = sigalrm;
		sigfillset(&act.sa_mask);
		sigaction(SIGALRM, &act, NULL);
		char *text = "Neque porro quisquam est qui dolorem ipsum quia dolor sit amet, consectetur, adipisci velit...%d";
		if(cflag){
			for(i = 0; i < 10; i ++){
				int n;
				n = snprintf(buf, sizeof buf, text, i);
				write(fd, buf, n);
				sleep(1);
			}
		} else {
			int n;
			alarm(15);
			for(i = 0; i < 10; i ++){
				if((n = read(fd, buf, sizeof buf)) <= 0)
					error("read");
				if(write(1, buf, n) == -1)
					error("write");
				if(write(1, "\n", 1) == -1)
					error("write");
			}
			alarm(0);
		}
	}

	close(fd);
	return 0;
}

